# -*- coding: utf-8 -*-
#
# web_container.py
#
# Copyright © 2016-2017 Antergos
#
# This file is part of whither.
#
# whither is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# whither is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The following additional terms are in effect as per Section 7 of the license:
#
# The preservation of all legal notices and author attributions in
# the material or in the Appropriate Legal Notices displayed
# by works containing it is required.
#
# You should have received a copy of the GNU General Public License
# along with whither; If not, see <http://www.gnu.org/licenses/>.

# Standard Lib
import os
from typing import (
    Tuple,
    TypeVar,
)

# 3rd-Party Libs
from PyQt5.QtWebChannel import QWebChannel
from PyQt5.QtWebEngineWidgets import (
    QWebEnginePage,
    QWebEngineView,
    QWebEngineSettings,
    QWebEngineScript,
)
from PyQt5.QtCore import (
    QUrl,
    QFile,
)

# This Library
from whither.base.objects import WebContainer
from .devtools import DevTools

# Typing Helpers
BridgeObjects = Tuple['BridgeObject']
Url = TypeVar('Url', str, QUrl)


DEFAULT_ENTRY_POINT = 'Whither: Unable to load entry point.'

DISABLED_SETTINGS = [
    'PluginsEnabled',  # Qt 5.6+
]

ENABLED_SETTINGS = [
    'FocusOnNavigationEnabled',  # Qt 5.8+
    'FullScreenSupportEnabled',  # Qt 5.6+
    'LocalContentCanAccessFileUrls',
    'ScreenCaptureEnabled',      # Qt 5.7+
    'ScrollAnimatorEnabled',
]


class QtWebContainer(WebContainer):

    def __init__(self, bridge_objs: BridgeObjects = None, *args, **kwargs) -> None:
        super().__init__(name='_web_container', bridge_objs=bridge_objs, *args, **kwargs)

        if self._config.debug_mode:
            os.environ['QTWEBENGINE_REMOTE_DEBUGGING'] = '12345'

        self.view = QWebEngineView(parent=self._main_window.widget)
        self.page = self.view.page()
        self.channel = QWebChannel(self.page)
        self.bridge_initialized = False

        self._initialize_page(self.page)

        if self._config.debug_mode:
            self.devtools = DevTools()

        if self._config.entry_point.autoload:
            self.initialize_bridge_objects()
            self.load()

        self.view.show()
        self._main_window.widget.setCentralWidget(self.view)

    @staticmethod
    def _create_webengine_script(path: Url, name: str) -> QWebEngineScript:
        script = QWebEngineScript()
        script_file = QFile(path)

        if script_file.open(QFile.ReadOnly):
            script_string = str(script_file.readAll(), 'utf-8')

            script.setInjectionPoint(QWebEngineScript.DocumentCreation)
            script.setName(name)
            script.setWorldId(QWebEngineScript.MainWorld)
            script.setSourceCode(script_string)

        return script

    def _get_channel_api_script(self) -> QWebEngineScript:
        return self._create_webengine_script(':/qtwebchannel/qwebchannel.js', 'QWebChannel API')

    def _init_bridge_channel(self) -> None:
        self.page.setWebChannel(self.channel)
        self.page.scripts().insert(self._get_channel_api_script())

        self.bridge_initialized = True

    def _initialize_page(self, page: QWebEnginePage) -> None:
        page_settings = self.page.settings().globalSettings()

        if self._config.allow_remote_urls:
            ENABLED_SETTINGS.append('LocalContentCanAccessRemoteUrls')
        else:
            DISABLED_SETTINGS.append('LocalContentCanAccessRemoteUrls')

        for setting in DISABLED_SETTINGS:
            try:
                page_settings.setAttribute(getattr(QWebEngineSettings, setting), False)
            except AttributeError:
                pass

        for setting in ENABLED_SETTINGS:
            try:
                page_settings.setAttribute(getattr(QWebEngineSettings, setting), True)
            except AttributeError:
                pass

        page.setView(self.view)

    def initialize_bridge_objects(self) -> None:
        if not self.bridge_initialized:
            self._init_bridge_channel()

        registered_objects = self.channel.registeredObjects()

        for obj in self.bridge_objects:
            if obj not in registered_objects:
                self.channel.registerObject(obj._name, obj)

    def load(self, url: str = '') -> None:
        url = url if url else self._config.entry_point.url

        if 'http' not in url and not os.path.exists(url):
            self.page.setHtml(DEFAULT_ENTRY_POINT)
            return

        if not url.startswith('file'):
            url = f'file://{url}'

        self.page.load(QUrl(url))

    def load_script(self, path: Url, name: str):
        script = self._create_webengine_script(path, name)
        return self.page.scripts().insert(script)



# -*- coding: utf-8 -*-
#
# interceptor.py
#
# Copyright © 2016-2017 Antergos
#
# This file is part of whither.
#
# whither is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# whither is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The following additional terms are in effect as per Section 7 of the license:
#
# The preservation of all legal notices and author attributions in
# the material or in the Appropriate Legal Notices displayed
# by works containing it is required.
#
# You should have received a copy of the GNU General Public License
# along with whither; If not, see <http://www.gnu.org/licenses/>.

""" Wrapper for QMainWindow """

# 3rd-Party Libs
from PyQt5.QtCore import QObject
from PyQt5.QtWebEngineCore import (
    QWebEngineUrlRequestInterceptor,
    QWebEngineUrlRequestInfo,
)


class QtUrlRequestInterceptor(QWebEngineUrlRequestInterceptor):

    def intercept_request(self, info: QWebEngineUrlRequestInfo) -> None:
        pass

    def interceptRequest(self, info: QWebEngineUrlRequestInfo) -> None:
        self.intercept_request(info)



# -*- coding: utf-8 -*-
#
# app.py
#
# Copyright © 2016-2017 Antergos
#
# This file is part of whither.
#
# whither is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# whither is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The following additional terms are in effect as per Section 7 of the license:
#
# The preservation of all legal notices and author attributions in
# the material or in the Appropriate Legal Notices displayed
# by works containing it is required.
#
# You should have received a copy of the GNU General Public License
# along with whither; If not, see <http://www.gnu.org/licenses/>.

""" The primary entry point to the library. """

# Standard Lib
import os
import subprocess

# This Library
from .toolkits.bootstrap import Application, Window, WebContainer
from .base.objects import BridgeObjects
from .base.config_loader import ConfigLoader


class App(Application):

    def __init__(self, app_name, bridge_objects: BridgeObjects = None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._wh_load_config(app_name)
        self._maybe_start_accessibility_service()

        self._before_main_window_init()
        Window('_main_window')

        self._before_web_container_init()
        WebContainer(bridge_objects=bridge_objects)

        self._main_window.show()

    def _maybe_start_accessibility_service(self):
        if not self._config.at_spi_service.enabled:
            return

        if os.path.exists(self._config.at_spi_service.command):
            subprocess.run([self._config.at_spi_service.command, self._config.at_spi_service.arg])

    def _wh_load_config(self, key: str) -> None:
        try:
            setattr(self, 'config', ConfigLoader(key, '__main__').config['app'])
            setattr(self, '_config', ConfigLoader(key, '__main__').config['whither'])
        except Exception as err:
            setattr(self, 'config', ConfigLoader(key, __file__).config['app'])
            setattr(self, '_config', ConfigLoader(key, __file__).config['whither'])

    def _before_web_container_init(self):
        pass

    def _before_main_window_init(self):
        pass


if __name__ == '__main__':
    raise RuntimeError('Whither is a library (this module must be imported!)')



# -*- coding: utf-8 -*-
#
# object.py
#
# Copyright © 2016-2017 Antergos
#
# This file is part of whither.
#
# whither is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# whither is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The following additional terms are in effect as per Section 7 of the license:
#
# The preservation of all legal notices and author attributions in
# the material or in the Appropriate Legal Notices displayed
# by works containing it is required.
#
# You should have received a copy of the GNU General Public License
# along with whither; If not, see <http://www.gnu.org/licenses/>.

""" Data descriptor objects to make data/settings easily accessible in all child classes. """

from threading import RLock


class AttributeDict(dict):
    """
    Dict subclass which provides access to its keys' values as attributes.

    See:
        dict.__doc__()

    """
    def __init__(self, seq=None, **kwargs):
        _kwargs = kwargs

        if isinstance(seq, dict):
            seq = self._process_dict(seq)
            _kwargs = dict(seq, **kwargs)

        super().__init__(**_kwargs)

        self['_lock'] = RLock()

    def __getattr__(self, attr):
        if attr in self:
            return self[attr]

        return None

    def __setattr__(self, attr, value):
        return self.__setitem__(attr, value)

    def __setitem__(self, item, value):
        if '_lock' == item:
            return super().__setitem__(item, value)

        value = self.maybe_make_attribute_dict(value)

        with self._lock:
            return super().__setitem__(item, value)

    @staticmethod
    def _process_dict(from_dict):
        for key, value in from_dict.items():
            if isinstance(value, dict) and not isinstance(value, AttributeDict):
                from_dict[key] = AttributeDict(value)

        return from_dict

    def as_dict(self):
        return {k: v for k, v in self.items()}

    @staticmethod
    def maybe_make_attribute_dict(value):
        if not isinstance(value, dict) or isinstance(value, AttributeDict):
            return value

        return AttributeDict(value)


class SharedData:
    """
    Descriptor that facilitates shared data storage/retrieval.

    Attributes:
        name      (str):  The name of the bound attribute.
        from_dict (dict): Initial data to store.

    """
    _data = {}

    def __init__(self, name: str = None, from_dict: dict = None) -> None:
        self.name = name

        if name is not None and from_dict is not None:
            self._data[name] = AttributeDict(from_dict)

    def __get__(self, instance, cls):
        if instance is None:
            return self

        if self.name not in self._data:
            self._data[self.name] = None
            return None

        return self._data[self.name]

    def __set__(self, instance, value):
        value = AttributeDict.maybe_make_attribute_dict(value)
        self._data[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name



# -*- coding: utf-8 -*-
#
# window.py
#
# Copyright © 2016-2017 Antergos
#
# This file is part of whither.
#
# whither is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# whither is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The following additional terms are in effect as per Section 7 of the license:
#
# The preservation of all legal notices and author attributions in
# the material or in the Appropriate Legal Notices displayed
# by works containing it is required.
#
# You should have received a copy of the GNU General Public License
# along with whither; If not, see <http://www.gnu.org/licenses/>.

""" Wrapper for QMainWindow """

# Standard Lib
from typing import Dict

# 3rd-Party Libs
from PyQt5.QtWidgets import (
    QMainWindow,
    QAction,
    qApp,
    QWidget,
)
from PyQt5.QtCore import (
    QEvent,
    Qt,
)
from PyQt5.QtGui import QIcon

# This Library
from whither.base.objects import Window

# Typing Helpers
QtWindowStatesT = Dict[str, Qt.WindowState]


WINDOW_STATES: QtWindowStatesT = {
    'NORMAL': Qt.WindowNoState,
    'MINIMIZED': Qt.WindowMinimized,
    'MAXIMIZED': Qt.WindowMaximized,
    'FULLSCREEN': Qt.WindowFullScreen,
}


class QtWindow(Window):

    def __init__(self, name: str = '_window', *args, **kwargs) -> None:
        super().__init__(name=name, *args, **kwargs)

        self.states = WINDOW_STATES  # type: dict
        self.widget = None           # type: QWidget

        self._initialize()

    def _initialize(self) -> None:
        config, toolbar_config, initial_state = super()._initialize()

        if not self._app.windows:
            self.widget = QMainWindow()
        else:
            self.widget = QWidget()

        self._app.windows.append(self.widget)

        self.widget.setAttribute(Qt.WA_DeleteOnClose)

        if config.title:
            self.widget.setWindowTitle(config.title)

        if config.icon:
            self.widget.setWindowIcon(QIcon(config.icon))

        if config.width and config.height:
            self.widget.setFixedSize(config.width, config.height)

        if config.decorated and toolbar_config.enabled:
            self._init_menu_bar()

        elif not config.decorated:
            self.widget.setWindowFlags(self.widget.windowFlags() | Qt.FramelessWindowHint)

        if config.stays_on_top:
            self.widget.setWindowFlags(
                self.widget.windowFlags() | Qt.WindowStaysOnTopHint
            )

        self.widget.setWindowFlags(
            self.widget.windowFlags() | Qt.MaximizeUsingFullscreenGeometryHint
        )

        self.set_state(self.states[initial_state])

    def _init_menu_bar(self) -> None:
        exit_action = QAction(QIcon('exit.png'), '&Exit', self)
        exit_action.setShortcut('Ctrl+Q')
        exit_action.setStatusTip('Exit application')
        exit_action.triggered.connect(qApp.quit)

        menu_bar = self.widget.menuBar()

        file_menu = menu_bar.addMenu('&File')
        file_menu.addAction(exit_action)

        edit_menu = menu_bar.addMenu('&Edit')
        edit_menu.addAction(exit_action)

        view_menu = menu_bar.addMenu('&View')
        view_menu.addAction(exit_action)

        about_menu = menu_bar.addMenu('&About')
        about_menu.addAction(exit_action)

    def changeEvent(self, event: QEvent) -> None:
        if event is not QEvent.WindowStateChange:
            return

        try:
            new_window_state = self.widget.windowHandle().windowState()
        except Exception:
            new_window_state = self.widget.windowState()

        if new_window_state & self.states['MAXIMIZED']:
            self.state = self.states['MAXIMIZED']

        elif new_window_state & self.states['FULLSCREEN']:
            self.state = self.states['FULLSCREEN']

        elif new_window_state & self.states['MINIMIZED']:
            self.state = self.states['MINIMIZED']

        else:
            self.state = self.states['NORMAL']

    def show(self) -> None:
        self.widget.show()

    def show_fullscreen(self) -> None:
        try:
            self.widget.windowHandle().showFullScreen()
        except Exception:
            self.widget.showFullScreen()

    def show_maximized(self) -> None:
        try:
            self.widget.windowHandle().showMaximized()
        except Exception:
            self.widget.showMaximized()

    def show_minimized(self) -> None:
        try:
            self.widget.windowHandle().showMinimized()
        except Exception:
            self.widget.showMinimized()

    def set_state(self, new_state: Qt.WindowState) -> None:
        if new_state == self.state:
            return

        try:
            self.widget.windowHandle().setWindowState(new_state)
        except Exception:
            self.widget.setWindowState(new_state)




# -*- coding: utf-8 -*-
#
# application.py
#
# Copyright © 2016-2017 Antergos
#
# This file is part of whither.
#
# whither is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# whither is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The following additional terms are in effect as per Section 7 of the license:
#
# The preservation of all legal notices and author attributions in
# the material or in the Appropriate Legal Notices displayed
# by works containing it is required.
#
# You should have received a copy of the GNU General Public License
# along with whither; If not, see <http://www.gnu.org/licenses/>.

""" Wrapper for QMainWindow """

# 3rd-Party Libs
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication

# This Lib
from whither.base.objects import Application


class QtApplication(Application):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(name='_app', *args, **kwargs)

        self.widget = QApplication([])
        self.is_qt, self.is_gtk = True, False
        self.desktop = self.widget.desktop()

        self.widget.setAttribute(Qt.AA_EnableHighDpiScaling)

    def _set_window_size_position(self) -> None:
        if self._config.window.no_desktop_env is False:
            return

        self._main_window.widget.setGeometry(self.desktop.availableGeometry())

    def run(self) -> int:
        self._set_window_size_position()

        return self.widget.exec_()



# -*- coding: utf-8 -*-
#
# objects.py
#
# Copyright © 2016-2017 Antergos
#
# This file is part of whither.
#
# whither is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# whither is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The following additional terms are in effect as per Section 7 of the license:
#
# The preservation of all legal notices and author attributions in
# the material or in the Appropriate Legal Notices displayed
# by works containing it is required.
#
# You should have received a copy of the GNU General Public License
# along with whither; If not, see <http://www.gnu.org/licenses/>.

""" Bases Classes (for future flexibility) """

# Standard Lib
from typing import (
    ClassVar,
    Tuple,
    Type,
    TypeVar,
)

# This Lib
from .data import (
    AttributeDict,
    SharedData,
)

# Typing Helpers
BridgeObjects = Tuple[Type['BridgeObjectBase']]
Widget = TypeVar('Widget', 'QObject', 'GObject')

# Application class is a singleton
_APP_INSTANCE = None

_MAIN_COMPONENTS = (
    '_app',
    '_config',
    '_logger',
    '_main_window',
    '_web_container',
    'config',
)


class BaseObject:
    _app = SharedData()            # type: ClassVar[Type['Application']]
    _config = SharedData()         # type: ClassVar[Type[AttributeDict]]
    _logger = SharedData()         # type: ClassVar[Type['Logger']]
    _main_window = SharedData()    # type: ClassVar[Type['Window']]
    _web_container = SharedData()  # type: ClassVar[Type['WebContainer']]

    config = SharedData()          # type: ClassVar[Type[AttributeDict]]
    is_gtk = None                  # type: ClassVar[bool]
    is_qt = None                   # type: ClassVar[bool]
    windows = SharedData()         # type: ClassVar[list]

    def __init__(self, name: str, *args, **kwargs) -> None:
        self.widget: Widget = None
        self.name: str = name

        if name in _MAIN_COMPONENTS:
            self._register_main_component(name)

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__()
        cls.__pre_init__()

    @classmethod
    def __pre_init__(cls):
        pass

    def _register_main_component(self, name: str) -> None:
        attrib = getattr(self, name)

        if attrib is None:
            setattr(self, name, self)


class Application(BaseObject):

    windows: ClassVar[list] = []

    def __init__(self, name: str, *args, **kwargs) -> None:
        global _APP_INSTANCE
        if _APP_INSTANCE is not None:
            return

        _APP_INSTANCE = self

        super().__init__(name=name, *args, **kwargs)

    def run(self) -> int:
        raise NotImplementedError()


class WebContainer(BaseObject):

    def __init__(self, name: str, bridge_objs: BridgeObjects = None, *args, **kwargs) -> None:
        super().__init__(name=name, *args, **kwargs)

        self.bridge_objects = bridge_objs or ()

    def initialize_bridge_objects(self) -> None:
        raise NotImplementedError()

    def load(self, url: str) -> None:
        raise NotImplementedError


class Window(BaseObject):

    states = SharedData()  # type: ClassVar[AttributeDict]

    def __init__(self, name: str, *args, **kwargs) -> None:
        super().__init__(name=name, *args, **kwargs)

        self.state: int = 0

    def _initialize(self) -> tuple:
        config = self._config.window
        toolbar_config = self._config.toolbar
        initial_state = config.initial_state.upper()

        return config, toolbar_config, initial_state

    def show(self) -> None:
        raise NotImplementedError()

    def set_state(self, state: int) -> None:
        raise NotImplementedError()



# Always prefer setuptools over distutils
from setuptools import setup, find_packages

setup(
    name='whither',
    version='0.2.1',
    packages=[
        'whither',
        'whither.base',
        'whither.toolkits',
        'whither.toolkits.qt',
        'whither.toolkits.gtk',
    ],
    url='https://github.com/antergos/whither',
    license='GPL-3.0',
    author='Antergos Linux Project',
    author_email='dustin@antergos.com',
    description='Desktop application SDK for creating Universal Linux Applications.',
    install_requires=[
        'PyQt5',
        'ruamel.yaml',
    ],
    package_data={
        '': ['whither.yml'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: X11 Applications',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: JavaScript',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: User Interfaces',
    ],
    keywords='desktop-application-sdk framework sdk javascript universal html5'
)



# -*- coding: utf-8 -*-
#
# window.py
#
# Copyright © 2016-2017 Antergos
#
# This file is part of whither.
#
# whither is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# whither is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The following additional terms are in effect as per Section 7 of the license:
#
# The preservation of all legal notices and author attributions in
# the material or in the Appropriate Legal Notices displayed
# by works containing it is required.
#
# You should have received a copy of the GNU General Public License
# along with whither; If not, see <http://www.gnu.org/licenses/>.

""" Wrapper for GtkWindow """

# Standard Lib

# 3rd-Party Libs
import gi
gi.require_versions({'Gtk': '3.0', 'Gdk': '3.0'})
from gi.repository import (
    Gdk,
    Gtk,
)

# This Library
from whither.base.objects import Window


WINDOW_STATES = {
    'NORMAL': 0,
    'MINIMIZED': Gdk.WindowState.ICONIFIED,
    'MAXIMIZED': Gdk.WindowState.MAXIMIZED,
    'FULLSCREEN': Gdk.WindowState.FULLSCREEN,
}


class GtkWindow(Window):

    def __init__(self, name: str = '_window', *args, **kwargs) -> None:
        super().__init__(name=name, *args, **kwargs)

        self.states = WINDOW_STATES  # type: dict
        self.widget = None           # type: Gtk.Window

    def _initialize(self) -> None:
        config, toolbar_config, initial_state = super()._initialize()

        if not self._app.windows:
            self.widget = Gtk.ApplicationWindow()
        else:
            self.widget = Gtk.Window()

        self._app.windows.append(self.widget)

        if config.title:
            self.widget.set_title(config.title)

        if config.icon:
            self.widget.set_default_icon_from_file(config.icon)

        if config.width and config.height:
            self.widget.set_size_request(config.width, config.height)

        if config.decorated and toolbar_config.enabled:
            self._init_menu_bar()

        elif not config.decorated:
            self.widget.set_decorated(False)

        if config.stays_on_top:
            self.widget.set_keep_above(True)

        self.set_state(self.states[initial_state])

    def _init_menu_bar(self) -> None:
        pass

    def _set_state_normal(self):
        if self.state is self.states['MAXIMIZED']:
            self.widget.unmaximize()

        elif self.state is self.states['FULLSCREEN']:
            self.widget.unfullscreen()

        elif self.state is self.states['MINIMIZED']:
            self.widget.unfullscreen()

    def _window_state_event_cb(self, window, event, *args):
        if event.new_window_state & self.states['MAXIMIZED']:
            self.state = self.states['MAXIMIZED']

        elif event.new_window_state & self.states['FULLSCREEN']:
            self.state = self.states['FULLSCREEN']

        elif event.new_window_state & self.states['MINIMIZED']:
            self.state = self.states['MINIMIZED']

        else:
            self.state = 0

    def show(self) -> None:
        self.widget.show_all()

    def show_fullscreen(self) -> None:
        self.widget.fullscreen()
        self.widget.show_all()

    def show_maximized(self) -> None:
        self.widget.maximize()
        self.widget.show_all()

    def show_minimized(self) -> None:
        self.widget.iconify()
        self.widget.show_all()

    def set_state(self, new_state: int) -> None:
        if new_state == self.state:
            return

        if new_state is self.states['NORMAL']:
            self._set_state_normal()

        elif new_state is self.states['MAXIMIZED']:
            self.widget.maximize()

        elif new_state is self.states['FULLSCREEN']:
            self.widget.fullscreen()

        elif new_state is self.states['MINIMIZED']:
            self.widget.iconify()





# -*- coding: utf-8 -*-
#
# config.py
#
# Copyright © 2016-2017 Antergos
#
# This file is part of whither.
#
# whither is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# whither is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The following additional terms are in effect as per Section 7 of the license:
#
# The preservation of all legal notices and author attributions in
# the material or in the Appropriate Legal Notices displayed
# by works containing it is required.
#
# You should have received a copy of the GNU General Public License
# along with whither; If not, see <http://www.gnu.org/licenses/>.

# Standard Lib
import pkg_resources
from typing import Callable

# 3rd Party Libs
import ruamel.yaml as yaml


class ConfigLoader:
    _filters = []

    def __init__(self, key: str, path: str = None) -> None:
        self.load_from = path
        self.config = self.load_config(key)

    def _filter_data(self, key: str, data: str) -> str:
        if not self._filters:
            return data

        for callback in self._filters:
            data = callback(key, data)

        return data

    @classmethod
    def add_filter(cls, callback: Callable[[str, str], None]) -> None:
        cls._filters.append(callback)

    def load_config(self, key: str) -> dict:
        try:
            data = pkg_resources.resource_string(self.load_from, 'whither.yml').decode('utf-8')
            data = self._filter_data(key, data)
            config = yaml.safe_load(data)
            config = config[key]
        except Exception:
            data = open(self.load_from, 'r').read()
            data = self._filter_data(key, data)
            config = yaml.safe_load(data)
            config = config[key]

        return {key: value for key, value in config.items()}

    @classmethod
    def remove_filter(cls, callback: Callable[[str, str], None]) -> None:
        cls._filters.remove(callback)
